/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "resampler.hh"
#include "DVecType.hh"
#include "gcd.hh"
#include "FIRdesign.hh"
#include "FIRdft.hh"
#include "pipe_tools.hh"
#include <stdexcept>
#include <iostream>

using namespace std;

//=======================================  Resample default constructor
resampler::resampler(void)
   : sample_rate(0), edgewid(0.02), attenuation(60),
     firmode(FIRdft::fm_zero_phase), defacto(0), startTime(0),
     currentTime(0), inStep(0)
{
}

//=======================================  Resample initializing constructor
resampler::resampler(double to_rate)
   : firmode(FIRdft::fm_zero_phase), defacto(0), startTime(0),
     currentTime(0), inStep(0)
{
   set_sample_rate(to_rate);
   set_AA_parameters(0.02, 60.0);
}

//=======================================  Resample destructor
resampler::~resampler(void) {
}

//=======================================  Clone this resampler
resampler*
resampler::clone(void) const {
   return new resampler(*this);
}

//=======================================  Resample a specified time series
TSeries
resampler::apply(const TSeries& in) {
   //------------------------------------  set up on first use
   if (!inUse()) {
      startTime = in.getStartTime();
      currentTime = startTime;
      inStep = in.getTStep();
      setup();
   }
  
   //------------------------------------  Otherwise, check for valid input
   else {
      dataCheck(in);
   }
   currentTime = in.getEndTime(); 

   //------------------------------------  start with interpolation
   TSeries tin;
   if (interpolate.null()) {
      tin = in;
   } else {
      tin = interpolate(in);
   }

   //------------------------------------  followed by decimater
   if (defacto <= 1) return tin;
   TSeries aaSeries(antiAlias(tin));
   if (aabuffer.Append(aaSeries)) {
      cerr << "resampler: aaSeries start: " << aaSeries.getStartTime()
	   << " aaBuffer end: " << aabuffer.getEndTime()
	   << " input series start: " << in.getStartTime() << endl;
      throw runtime_error("resampler: antialiasing filter alignment error");
   }
   int ndecim = aabuffer.getNSample() / defacto;
   Interval out_len = aabuffer.getTStep() * double(ndecim * defacto);
   TSeries r(aabuffer.decimate(defacto));
   aabuffer.eraseStart(out_len);
   return r;
}

//======================================  Check for valid input data
void 
resampler::dataCheck(const TSeries& in) const {
   if (inUse()) {
      if (in.getTStep() != inStep) {
	 throw runtime_error("resampler: input sample rate has changed.");
      } else if (in.getStartTime() != currentTime) {
	 cerr << "resampler: input start time: " << in.getStartTime()
	      << " current time: " << currentTime << endl;
	 throw runtime_error("resampler: input series not continuous.");
      }
   }
}

//====================================  Time delay
Interval
resampler::getTimeDelay(void) const {
   return antiAlias.null() ? Interval(0) : antiAlias->getTimeDelay();
}

//====================================  reset the filter
void 
resampler::reset(void) {
   currentTime = startTime = Time(0);
   interpolate.reset();
   antiAlias.reset();
   aabuffer.Clear();
}

//======================================  Set anti-aliasing filter parameters
void
resampler::set_AA_parameters(double edge, double atten_dB) {
   edgewid = edge;
   attenuation = atten_dB;
}

//======================================  Set anti-aliasing filter parameters
void
resampler::setMode(int fmode) {
   firmode = fmode;
   set_fir_mode(antiAlias.get(), firmode);
}

//======================================  set target sample rate
void 
resampler::set_sample_rate(double to_rate) {
   sample_rate = to_rate;
   inStep = 0;
   reset();
}

//======================================  Design the antialiasing filter.
void 
resampler::setup(void) {
   if (!sample_rate) {
      throw runtime_error("resampler: target rate not set.");
   } else if (!inStep) {
      throw runtime_error("resampler: input sample rate unknown.");
   }
   defacto = int(1./double(sample_rate * inStep));
   if (sample_rate * defacto * double(inStep) != 1) {
      throw runtime_error("resampler: non-integer decimation factor.");
   }

   //--------------------------------  Design an anti-aliasing filter.
   DVectD filter_coefs;
   double edge = (0.5 - edgewid) / defacto;
   double df   = 2.0 * edgewid / defacto;
   int ncof = 0;
   double* h = 0;
   int stat = FirW(ncof, 6, 1, edge, 0.0, df, attenuation, h);
   if (stat > 0) {
      //--------------------------  Preserve the decimated signal phase.
      //   To maintain the sample phasing, the delay must be an integer
      //   multiple of the decimated sample rate, so the filter order
      //   should <filter-order> = 2 * <multiplier> * <decimation-factor>
      //   Include a factor to force delay to an even number of nanoseconds. 
      long inHz(1. / inStep);
      long nsFact = inHz / gcd(1000000000, inHz);
      long dtFact = nsFact * defacto / gcd(nsFact, defacto);
      if ((ncof - 1) % (2 * dtFact) != 0) {
	 int mult = (ncof - 1) / (2 * dtFact) + 1;
	 ncof = mult * 2 * dtFact + 1;
      }
      
      //--------------------------  Allocate the coeff. array and design
      filter_coefs.ReSize(ncof);
      h = filter_coefs.refTData();
      stat = FirW(ncof, 6, 1, edge, 0.0, df, attenuation, h);
   }
   if (stat != 0) {
      throw runtime_error("single_resample::filter FirW failed"); 
   }
   cout << "resampler: Design FirW, order: " << ncof - 1 << endl;
   FIRdft* filt = new FIRdft(ncof - 1, 1.0 / inStep);
   filt->setCoefs(ncof, filter_coefs.refTData());
   filt->setMode(FIRdft::fir_mode(firmode));
   antiAlias.set(filt);
}
